' ----------------------------------------------------------------
' SIMPLE ZX BASIC LIBRARY FOR JOYSTICK CONTROLS - by Einar Saukas
' ----------------------------------------------------------------

#ifndef __LIBRARY_JOY__
#define __LIBRARY_JOY__
#pragma push(case_insensitive)
#pragma case_insensitive = true

' ----------------------------------------------------------------
' Constants to select active control (joystick or keyboard)
' ----------------------------------------------------------------
#define JOY_KEYBOARD  0
#define JOY_KEMPSTON  1
#define JOY_SINCLAIR  2

' ----------------------------------------------------------------
' Constants to check each movement direction
' ----------------------------------------------------------------
#define JOY_FIRE   1
#define JOY_UP     2
#define JOY_DOWN   4
#define JOY_RIGHT  8
#define JOY_LEFT  16

' ----------------------------------------------------------------
' Read controls (Left/Right/Down/Up/Fire) from Sinclair Joystick
' ----------------------------------------------------------------
function joySinclair() AS UBYTE
    asm
        ld a, $ef
        in a, ($fe)     ; A = ???67890
        cpl
        and $1f
    end asm
end function

' ----------------------------------------------------------------
' Read controls Left/Right/Down/Up/Fire from Keyboard
' ----------------------------------------------------------------
function joyKeyboard() AS UBYTE
    asm
        ld a, $df
        in a, ($fe)     ; read O/P
        ld c, a         ; C = ??????OP
        ld a, $fd
        in a, ($fe)     ; read A
        rra
        rl c            ; C = ?????OPA
        ld a, $fb
        in a, ($fe)     ; read Q
        rra
        rl c            ; C = ????OPAQ
        ld a, $7f
        in a, ($fe)     ; read M/SPACE
        and 5
        add a, 251
        ld a,c
        rla             ; A = ???OPAQM
        cpl
        and $1f
    end asm
end function

' ----------------------------------------------------------------
' Read controls Left/Right/Down/Up/Fire from Kempston Joystick
' ----------------------------------------------------------------
function joyKempston() AS UBYTE
    asm
        in a,($1f)      ; A = ???FUDLR
        ld c,a          ; C = ???FUDLR
        rra
        rra
        rra
        rl c            ; C = ??FUDLRD
        rra
        rl c            ; C = ?FUDLRDU
        rra
        rl c            ; C = FUDLRDUF
        ld a,c          ; A = ???LRDUF
        and $1f
    end asm
end function

' ----------------------------------------------------------------
' Read controls Left/Right/Down/Up/Fire from joystick or keyboard
' ----------------------------------------------------------------
function joyControl(joystick AS UBYTE) AS UBYTE
    IF joystick = JOY_SINCLAIR THEN
        RETURN joySinclair()
    ELSEIF joystick = JOY_KEMPSTON THEN
        RETURN joyKempston()
    ELSE
        RETURN joyKeyboard()
    END IF
end function

' ----------------------------------------------------------------
' THE END
' ----------------------------------------------------------------
#pragma pop(case_insensitive)
#endif
