' ------------------------------------------------------------------------------
' "ZEN"
'
' GAME BY EINAR SAUKAS
' MUSIC BY MISTER BEEP
' ------------------------------------------------------------------------------

' ------------------------------------------------------------------------------
' GLOBAL VARIABLES
' ------------------------------------------------------------------------------

    DIM joystick AS UBYTE                               ' selected joystick
    DIM width AS UBYTE                                  ' board width
    DIM height AS UBYTE                                 ' board height
    DIM row0 AS UBYTE                                   ' initial board row on screen
    DIM col0 AS UBYTE                                   ' initial board column on screen

    GO TO 10


' ------------------------------------------------------------------------------
' GAME DATA
' ------------------------------------------------------------------------------

'Compressed menu screen
menuscr:
    asm
        incbin "menu.scr.rcs.zx7"
    end asm

'Graphics font
zenchr:
    asm
        incbin "zen.chr"
    end asm

'Digits font
zen64chr:
    asm
        incbin "zen64.chr"
    end asm


' ------------------------------------------------------------------------------
' GAME CODE
' ------------------------------------------------------------------------------

#include "zx7.bas"
#include "joy.bas"

'Constants for tile connections
#define  CONNECT_RIGHT  1
#define  CONNECT_DOWN   2
#define  CONNECT_LEFT   4
#define  CONNECT_UP     8

'Constants for playing board
#define  BOARD_ADDR  32484
#define  BOARD(r,c)  (BOARD_ADDR+((c)<<4)+(r))


' ------------------------------------------------------------------------------
' Play specified music track until any key is pressed.
'
' Parameters:
'     track: music track number
' ------------------------------------------------------------------------------
sub FASTCALL playMusic(track AS UBYTE)
    asm
        dec     a
        jp      m, 35000                ; track 0: "Chromospheric Flares"
        jp      z, 40221                ; track 1: "Dragon Hammer"
        dec     a
        jp      z, 41429                ; track 2: "White Hot Iron"
        dec     a
        jp      z, 52805                ; track 3: "Galactic Coursing"
        jp      59460                   ; default: "Spiral Space Part Z"
    end asm
end sub


' ------------------------------------------------------------------------------
' Fill entire screen with specified color attribute.
'
' Parameters:
'     attrib: color attribute (0-255)
' ------------------------------------------------------------------------------
sub FASTCALL paintScreen(attrib AS UBYTE)
    asm
        halt                            ; wait 5 frames
        halt
        halt
        halt
        halt
        ld      hl, 22528               ; fill attribute area with value A
        ld      (hl), a
        ld      de, 22529
        ld      bc, 767
        ldir
    end asm
end sub

' ------------------------------------------------------------------------------
' Just like PAUSE in Sinclair BASIC, except pause cannot be interrupted by
' pressed keys.
'
' Parameters:
'     frames: number of frames to pause (1-255)
' ------------------------------------------------------------------------------
sub pauseFrames(frames as UBYTE)
    DO
        asm
            halt
        end asm
        LET frames = frames-1
    LOOP UNTIL frames=0
end sub

' ------------------------------------------------------------------------------
' Draw specified number digit at top right corner of playing board.
'
' Parameters:
'     pos: digit position (0-3)
'     digit: digit value (0-9)
' ------------------------------------------------------------------------------
sub drawDigit(pos AS UBYTE, digit AS UBYTE)
    DIM col AS UBYTE

    LET col = col0+(width<<1)+(pos>>1)
    IF pos & 1 THEN
        LET digit = digit+20
    END IF
    POKE UINTEGER 23606, @zen64chr-256
    PRINT OVER 1;AT row0, col;CHR$(digit+32);AT row0+1,col;CHR$(digit+42);
end sub

' ------------------------------------------------------------------------------
' Draw specified tile (2x2 characters) at specified location within the playing
' board using specified INK color.
'
' Parameters:
'     row: board row (0-11)
'     col: board column (0-15)
'     color: INK color
' ------------------------------------------------------------------------------
sub drawTile(row AS UBYTE, col AS UBYTE, color AS UBYTE)
    DIM tile AS UBYTE
    DIM ch AS UBYTE

    LET tile = PEEK BOARD(row, col)
    IF tile = 0 THEN
        IF row=0 AND col=width THEN
            OVER 1
        ELSEIF color=2 THEN
            LET tile = 16
        END IF
    END IF
    LET ch = (tile<<2)+32
    POKE UINTEGER 23606, @zenchr-256
    INK color
    PRINT AT row0+(row<<1), col0+(col<<1); CHR$(ch);CHR$(ch+1);AT row0+(row<<1)+1, col0+(col<<1);CHR$(ch+2);CHR$(ch+3);:
    INK 0
    OVER 0
end sub

' ------------------------------------------------------------------------------
' Count "open connections" in a certain tile at specified location within the
' playing board, that currently don't match any connections in their neighbor
' tiles.
'
' Parameters:
'     row: board row (0-11)
'     col: board column (0-15)
' ------------------------------------------------------------------------------
function openConnections(row AS UBYTE, col AS UBYTE) AS UINTEGER
    DIM tile AS UBYTE
    DIM n AS UINTEGER

    LET n = 0
    LET tile = PEEK BOARD(row, col)
    IF (tile & CONNECT_DOWN)  AND NOT (PEEK BOARD(row+1, col) & CONNECT_UP)    THEN LET n = n+1: END IF
    IF (tile & CONNECT_UP)    AND NOT (PEEK BOARD(row-1, col) & CONNECT_DOWN)  THEN LET n = n+1: END IF
    IF (tile & CONNECT_LEFT)  AND NOT (PEEK BOARD(row, col-1) & CONNECT_RIGHT) THEN LET n = n+1: END IF
    IF (tile & CONNECT_RIGHT) AND NOT (PEEK BOARD(row, col+1) & CONNECT_LEFT)  THEN LET n = n+1: END IF
    RETURN n
end function

' ------------------------------------------------------------------------------
' Main program
' ------------------------------------------------------------------------------
10  DIM level AS UINTEGER                               ' current level
    DIM counter AS UINTEGER                             ' total open connections
    DIM track AS UBYTE                                  ' current music track
    DIM tile AS UBYTE                                   ' current tile
    DIM row AS UBYTE                                    ' current cursor row within board
    DIM col AS UBYTE                                    ' current cursor column within board
    DIM row2 AS UBYTE                                   ' next cursor row within board
    DIM col2 AS UBYTE                                   ' next cursor column within board
    DIM b AS UBYTE                                      ' generic byte variable
    DIM i AS UINTEGER                                   ' generic integer variable

    'Play music at loading screen until any key is pressed
    BORDER 0
    WHILE INKEY$<>"" END WHILE
    IF PEEK 23733=255 THEN                              ' skip music in Spectrum 16K
        LET track = 7
        playMusic(track)
    ELSE
        WHILE INKEY$="" END WHILE
    END IF

    'Clear screen instantly
    paintScreen(0)

    DO
        'Set default screen colors
        INK 0: PAPER 0: FLASH 0: BRIGHT 0: CLS

        'Decompress/show menu screen
        dzx7AgileRCS(@menuscr, 16384)
        WHILE INKEY$<>"" END WHILE

        'Play music until joystick option is selected
        DO
            IF PEEK 23733=255 THEN                      ' skip music in Spectrum 16K
                WHILE INKEY$<>"" END WHILE
                LET track = (track+1) & 3
                PAUSE 1
                playMusic(track)
            END IF
            LET joystick = (CODE INKEY$)-(CODE "1")
        LOOP UNTIL joystick<3

        'Clear screen instantly
        paintScreen(0)

        'Play all levels
        FOR level = 0 TO 9999

            'Set randomize seed to ensure fixed layout for each level
            RANDOMIZE (level*5+1111)

            'Calculate playing board size depending on current level
            LET height = ((level+1)>>2)+2
            IF height>12 THEN height = 12: END IF
            LET width = ((level+3)>>2)+2
            IF width>16 THEN width = 16: END IF

            'Calculate playing board position on screen
            LET row0 = 12-height
            LET col0 = 14-width

            DO
                'Clear playing board
                FOR i = 0 TO 283: POKE BOARD_ADDR+i,0: NEXT i

                'Put tiles on playing board
                FOR row = 0 TO height-1
                    FOR col = width TO 1 STEP -1
                        IF row<>0 OR col<>width THEN
                            LET b = INT (RND * 256)
                            LET tile = ((PEEK BOARD(row, col+1) & CONNECT_LEFT) >> 2) | ((PEEK BOARD(row-1,col) & CONNECT_DOWN) << 2)
                            IF row <> height-1 THEN LET tile = tile | (b & CONNECT_DOWN) : END IF
                            IF col <> 1 THEN LET tile = tile | (b & CONNECT_LEFT) : END IF
                            IF b > 191 OR (b > 63 AND (row = 0 OR col = width)) OR tile = 0 THEN
                                b = b >> 2
                                IF row <> height-1 THEN LET tile = tile | (b & CONNECT_DOWN) : END IF
                                IF col <> 1 THEN LET tile = tile | (b & CONNECT_LEFT) : END IF
                            END IF
                            POKE BOARD(row, col), tile
                        END IF
                    NEXT col
                NEXT row

                'Rotate tiles, counting total number of open connections on board
                LET counter = 0
                FOR row = 0 TO height-1
                    FOR col = width TO 1 STEP -1
                        LET counter = counter - openConnections(row, col)
                        LET tile = PEEK BOARD(row, col)
                        FOR b = 1 TO INT(RND*4)
                            LET tile = ((tile << 1) & 0x0f) | (tile >> 3)
                        NEXT b
                        POKE BOARD(row, col), tile
                        LET counter = counter + openConnections(row, col)
                    NEXT col
                NEXT row

            'If algorithm generated an already solved board, then re-generate it
            LOOP UNTIL counter <> 0

            ' End previous level effect
            IF level<>0 THEN
                FOR b = 0 TO 6: paintScreen(6-b): NEXT b
            END IF
            CLS

            'Print level number at top right corner of the playing board
            drawDigit(0, INT(level/1000))
            drawDigit(1, INT(level/100) MOD 10)
            drawDigit(2, INT(level/10) MOD 10)
            drawDigit(3, level MOD 10)

            'Draw all tiles in the playing board
            FOR row = 0 TO height-1
                FOR col = width TO 1 STEP -1
                    drawTile(row, col, 0)
                NEXT col
            NEXT row

            'Change screen attributes to show playing board
            FOR b = 1 TO 6: paintScreen(b): NEXT b
            paintScreen(5)
            paintScreen(4)

            'Put cursor at top right corner of playing board
            LET row2 = 0
            LET col2 = width

            'Manage player controls
            LET b = JOY_FIRE
            DO
                'If cursor moved or rotated, then redraw current cursor location
                IF (b & JOY_FIRE) OR row2<>row OR col2<>col THEN
                    LET row = row2
                    LET col = col2
                    drawTile(row, col, 2)

                    'Wait until player releases controls
                    WHILE joyControl(joystick)<>0 END WHILE
                END IF

                'Read controls
                LET b = joyControl(joystick)

                'Move cursor
                IF (b & JOY_UP)    AND row>0        THEN LET row2 = row2-1: END IF
                IF (b & JOY_DOWN)  AND row<height-1 THEN LET row2 = row2+1: END IF
                IF (b & JOY_LEFT)  AND col>1        THEN LET col2 = col2-1: END IF
                IF (b & JOY_RIGHT) AND col<width    THEN LET col2 = col2+1: END IF

                'Rotate tile
                IF (b & JOY_FIRE) THEN
                    LET counter = counter - openConnections(row, col)
                    LET tile = PEEK BOARD(row,col)
                    POKE BOARD(row,col), ((tile << 1) & 0x0f) | (tile >> 3)
                    LET counter = counter + openConnections(row, col)
                END IF

                'If cursor moved or rotated, redraw last cursor position
                IF (b & JOY_FIRE) OR row2<>row OR col2<>col THEN
                    drawTile(row, col, 4)
                END IF

            'Repeat until level is solved
            LOOP UNTIL counter=0

            ' End level effect
            FOR b = 4 TO 7: paintScreen(b): NEXT b
            paintScreen(6)
            paintScreen(5)
            FOR b = 4 TO 7: paintScreen(b): NEXT b

        'Play next level
        NEXT level

        ' End last level effect
        pauseFrames(100): FOR b = 0 TO 6: paintScreen(6-b): pauseFrames(10): NEXT b
    LOOP

' ------------------------------------------------------------------------------
' THE END
' ------------------------------------------------------------------------------
